// https://qiita.com/tajima_taso/items/fdfed88c1e735ffb41e8

#include <stdio.h> //printf(), fprintf(), perror(), getc()
#include <stdlib.h> //strtod()
#include <sys/socket.h> //socket(), bind(), sendto(), recvfrom()
#include <arpa/inet.h> // struct sockaddr_in, struct sockaddr, inet_ntoa(), inet_aton()
#include <stdlib.h> //atoi(), exit(), EXIT_FAILURE, EXIT_SUCCESS
#include <string.h> //memset(), strcmp()
#include <unistd.h> //close()

#include <time.h>   //strftime(), localtime(), time()
#include <ctype.h>  //isspace()

#include "udp.h"

#include <pigpio.h>

// サーボのピン番号(BCM番号)
#define PIN_SERVO_1_STEER 6
#define PIN_SERVO_2_ACCEL 13
#define PIN_SERVO_3_LIGHT 19
#define PIN_SERVO_4_SHIFT 26

// http://edu.clipper.co.jp/pg-2-39.html
void trim( char *s ) {
    int i, j;
 
    //文字列の最後から空白を読み飛ばして除外する
    for( i = strlen(s)-1; i >= 0 && isspace( s[i] ); i-- ) ;
    s[i+1] = '\0';
    //先頭から空白でない文字まで読み飛ばす
    for( i = 0; isspace( s[i] ); i++ ) ;
    //前方の空白を詰める
    if( i > 0 ) {
        j = 0;
        while( s[i] ) s[j++] = s[i++];
        s[j] = '\0';
    }
}

void shift_change(int shift) {
    if (shift == 0)
        gpioServo(PIN_SERVO_4_SHIFT, 1500 + 400);
    else
        gpioServo(PIN_SERVO_4_SHIFT, 1500 + 400 - 400*(shift-1));
}

int main(int argc, char* argv[]) {

    const char *address = "";
    unsigned short port = (unsigned short)atoi(argv[1]);
    struct sockaddr servSockAddr, clitSockAddr;
    char recvBuffer[MAX_BUFSIZE];

    // UDPソケット初期化
    int server_sock = get_socket("udp");
    sockaddr_init(address, port, &servSockAddr);

    if (bind(server_sock, &servSockAddr, sizeof(servSockAddr)) < 0) {
        perror("bind() failed.");
        exit(EXIT_FAILURE);
    }

    // pigpio初期化
    if (gpioInitialise() < 0) {
        printf("\nGPIO cannot initialised\n");
        return 1;
    }
    gpioSetMode(PIN_SERVO_1_STEER, PI_OUTPUT);
    gpioSetMode(PIN_SERVO_2_ACCEL, PI_OUTPUT);
    gpioSetMode(PIN_SERVO_3_LIGHT, PI_OUTPUT);
    gpioSetMode(PIN_SERVO_4_SHIFT, PI_OUTPUT);

    gpioServo(PIN_SERVO_1_STEER, 1500);
    gpioServo(PIN_SERVO_2_ACCEL, 1500);
    gpioServo(PIN_SERVO_3_LIGHT, 1500);
    gpioServo(PIN_SERVO_4_SHIFT, 1500 + 400);

    // メインループ
    int shift = 1; // シフト(1, 2, 3, 0(バック))
    int brake = 0; // ブレーキ中=1
    int back  = 0; // バック中=1
    while(1) {
        char date[64];
        int recvMsgSize = udp_receive(server_sock, recvBuffer, MAX_BUFSIZE, &clitSockAddr);
        if (recvMsgSize == MSG_FAILURE) continue;

        // バッファの末尾に\0追加
        recvBuffer[recvMsgSize] = '\0';
        // バッファの前後の空白・改行を除去(trimモドキ)
        trim(recvBuffer);

        // タイムスタンプを文字列化
        // https://www.mm2d.net/main/prog/c/time-01.html
        time_t t = time(NULL);
        strftime(date, sizeof(date), "%Y/%m/%d %H:%M:%S", localtime(&t));

        // デバッグ出力
        printf("[%s] <%s> ", date,
            inet_ntoa(((struct sockaddr_in *)&clitSockAddr)->sin_addr));

        int sendMsgSize = udp_send(server_sock, "OK\n", &clitSockAddr);
        if (sendMsgSize == MSG_FAILURE) continue;

        // 受け取ったパラメータをパース
        char key[MAX_BUFSIZE];
        char val[MAX_BUFSIZE];
        //parse_param(recvBuffer, key, val);

        /*** keyの内容に応じて処理 ***/

        // ステアリング
        if (strncmp(recvBuffer, "STEERING=", 9) == 0) {
            // 浮動小数点(-1.0～+1.0)を読み取る
            double value = strtod(&recvBuffer[9], NULL);
            printf("STEERING    %f\n", value);
            gpioServo(PIN_SERVO_1_STEER, 1500 - 400 * value);
        }
        // アクセル
        else if (strncmp(recvBuffer, "ACCEL=", 6) == 0) {
            // 浮動小数点(0.0～+1.0)を読み取る
            double value = strtod(&recvBuffer[6], NULL);

            // 0.1以下はスキップ
            if (value < 0.1 && value != 0.0) {
                printf("\n");
                continue;
            }

            printf("ACCEL    %f\n", value);

            // バックの場合
            if (shift == 0) {
                // 符号を反転
                value = -value;

                // バックを離した瞬間
                if (value == 0.0) back = 0;
                // バックの踏み始めは一瞬バックを入れる
                else if (back == 0) {
                    back = 1;
                    gpioServo(PIN_SERVO_2_ACCEL, 1500 + 100);
                    gpioSleep(PI_TIME_RELATIVE, 0, 200000);
                    gpioServo(PIN_SERVO_2_ACCEL, 1500);
                    gpioSleep(PI_TIME_RELATIVE, 0, 10000);
                    printf("BACK    START!!\n");
                }
            }

            gpioServo(PIN_SERVO_2_ACCEL, 1500 - 400 * value);
        }
        // ブレーキ
        else if (strncmp(recvBuffer, "BRAKE=", 6) == 0) {
            // 浮動小数点(0.0～+1.0)を読み取る
            double value = strtod(&recvBuffer[6], NULL);

            // 0.1以下はスキップ
            if (value < 0.1 && value != 0.0) {
                printf("\n");
                continue;
            }

            printf("BRAKE    %f\n", value);

            // ブレーキを離した瞬間
            if (value == 0.0) brake = 0;
            // ブレーキの踏み始めは一瞬前進を入れる
            else if (brake == 0) {
                brake = 1;
                gpioServo(PIN_SERVO_2_ACCEL, 1500 - 50);
                gpioSleep(PI_TIME_RELATIVE, 0, 20000);
                gpioServo(PIN_SERVO_2_ACCEL, 1500);
                gpioSleep(PI_TIME_RELATIVE, 0, 10000);
                printf("BRAKE    START!!\n");
            }
            gpioServo(PIN_SERVO_2_ACCEL, 1500 + 400 * value);
        }

        // ホーン
        else if (strcmp(recvBuffer, "HORN=ON") == 0) {
            gpioServo(PIN_SERVO_3_LIGHT, 1500 - 400);
            printf("HORN    ON!!\n");
        }
        else if (strcmp(recvBuffer, "HORN=OFF") == 0) {
            gpioServo(PIN_SERVO_3_LIGHT, 1500);
            printf("HORN    OFF!\n");
        }
        // ハイビーム(バンパー前のフォグ)
        else if (strcmp(recvBuffer, "HIGHBEAM=ON") == 0) {
            gpioServo(PIN_SERVO_3_LIGHT, 1500 - 300);
            printf("HIGHBEAM    ON!!\n");
        }
        else if (strcmp(recvBuffer, "HIGHBEAM=OFF") == 0) {
            gpioServo(PIN_SERVO_3_LIGHT, 1500);
            printf("HIGHBEAM    OFF!\n");
        }
        // クラッチ

        // シフト
        else if (strcmp(recvBuffer, "SHIFT=UP") == 0) {
            shift++;
            if (shift > 3) shift = 3;
            shift_change(shift);
            printf("SHIFT   UP!!i   [%d]\n", shift);
        }
        else if (strcmp(recvBuffer, "SHIFT=DOWN") == 0) {
            shift--;
            if (shift < 0) shift = 0;
            shift_change(shift);
            printf("SHIFT   DOWN!!   [%d]\n", shift);
        }
        // ウィンカー(R2,L2)
        else if (strcmp(recvBuffer, "WINKER=RIGHT") == 0) {
            gpioServo(PIN_SERVO_1_STEER, 1500);
            gpioServo(PIN_SERVO_3_LIGHT, 1500 + 400);
            gpioSleep(PI_TIME_RELATIVE, 0, 100000); // sleep for 0.1 seconds
            gpioServo(PIN_SERVO_3_LIGHT, 1500);
            gpioSleep(PI_TIME_RELATIVE, 0, 100000); // sleep for 0.1 seconds
            gpioServo(PIN_SERVO_1_STEER, 1500 - 120);
            printf("WINKER  RIGHT\n");
        }
        else if (strcmp(recvBuffer, "WINKER=LEFT") == 0) {
            gpioServo(PIN_SERVO_1_STEER, 1500);
            gpioServo(PIN_SERVO_3_LIGHT, 1500 + 400);
            gpioSleep(PI_TIME_RELATIVE, 0, 100000); // sleep for 0.1 seconds
            gpioServo(PIN_SERVO_3_LIGHT, 1500);
            gpioSleep(PI_TIME_RELATIVE, 0, 100000); // sleep for 0.1 seconds
            gpioServo(PIN_SERVO_1_STEER, 1500 + 150);
            printf("WINKER  LEFT\n");
        }
        else if (strcmp(recvBuffer, "WINKER=OFF") == 0) {
            gpioServo(PIN_SERVO_1_STEER, 1500);
            printf("WINKER  OFF\n");
        }

        // ライト点灯(□)
        else if (strcmp(recvBuffer, "LIGHT=UP") == 0) {
            gpioServo(PIN_SERVO_1_STEER, 1500);
            gpioServo(PIN_SERVO_3_LIGHT, 1500 + 50 + 400);
            gpioSleep(PI_TIME_RELATIVE, 0, 100000); // sleep for 0.1 seconds
            gpioServo(PIN_SERVO_3_LIGHT, 1500 + 50);
            gpioSleep(PI_TIME_RELATIVE, 0, 100000); // sleep for 0.1 seconds
            printf("LIGHT   UP!\n");
        }
        // ハザード(△)
        else if (strcmp(recvBuffer, "HAZARD=ON") == 0) {
            gpioServo(PIN_SERVO_4_SHIFT, 1500 - 50);
            gpioSleep(PI_TIME_RELATIVE, 0, 500000); // sleep for 0.1 seconds
            gpioServo(PIN_SERVO_4_SHIFT, 1500 - 50 - 400);
            gpioSleep(PI_TIME_RELATIVE, 0, 500000); // sleep for 0.1 seconds
            gpioServo(PIN_SERVO_4_SHIFT, 1500 - 50);
            gpioSleep(PI_TIME_RELATIVE, 0, 500000); // sleep for 0.1 seconds
//            shift_change(shift);
            printf("HAZARD   ON!\n");
        }
        // エンジン停止・始動(PSボタン)


        // 未対応のパラメータ
        else {
            printf("not supported. '%s'\n", recvBuffer);
        }
    }

    // 終了処理
    gpioTerminate();
    return 0;
}

